/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


/*
** PROGRAM: analyze
** AUTHOR: David C. M. Ostapchuk
**
** DESCRIPTION:
**
** The program implements the statistical method to identify regions of
** differential variation in DNA sequences described in the paper:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** Given the results of a comparison of DNA sequences, this program computes:
**   - the T statistic;
**   - the G function;
**   - the Delta G function;
**   - the location of the peaks in G used to compute max(|Delta G|).
**
** These functions are defined and described in the bottom left corner on page
** 487 in the above paper.
*/


#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <cstring>  // strlen
#include <cstdlib>  // exit

#include "eventSequence.hh"
#include "io.hh"
#include "statistics.hh"
#include "misc.hh"

using namespace std;

const string version = "1.3";

void printVersion( void ) {
	cout << "analyze " << version << endl;
}


void printUsage( const char *executableName ) {
	cout << "Usage:\n"
	     << executableName
	     << " INPUT_FILENAME"
	     << " [OPTIONS]\n\n"
		 << "Analyzes the comparison sequences contained in INPUT_FILENAME. By default it\n"
	     << "calculates the T statistic and prints to it standard out.\n\n"
	     << "Options:\n"
	     << "    -T [T_FILENAME]           Output the T statistic to a file\n"
	     << "    -G [G_FILENAME]           Output the G function to a file\n"
	     << "    -D [DG_FILENAME]          Output the Delta G function to a file\n"
	     << "    -P [PEAKS_FILENAME]       Output the locations of the peaks in G to a file\n"
	     << "    -s SMOOTHING_PARAMETER    Use SMOOTHING_PARAMETER when detecting the\n"
	     << "                              peaks in G (the default is 0.005)\n"
	     << "    -d DELIMITER              Use DELIMITER to separate fields in the output\n"
	     << "                              files (the default is a comma)\n"
	     << "    -V                        Print the version number\n\n"
	     << "Default filenames:\n\n"
		 << "The output filenames are optional. The default filenames are generated by\n"
	     << "inserting \"_T\", \"_G\", \"_DG\", or \"_PEAKS\" at the end of INPUT_FILENAME,\n"
	     << "preserving the file extension if present."
	     << endl;
}


void bail( const char *executableName, const unsigned int errorCount ) {
	printUsage( executableName );
	exit( errorCount );
}


int main( int argc, char *argv[] ) {
	int precision = 10;                  // the max precision at which the data is output
	char delimiter = ',';                // default to CSV format
	double smoothingParameter = 0.005;   // default to what was used in Tang and Lewontin, Genetics 153:485, 1999

	vector< vector<double> > Gee, deltaGee;
	vector<double> Tee;
	vector< vector<unsigned int> > GPeaks;

	string inputFilename;
	string GFilename;
	string DGFilename;
	string TFilename;
	string GPeaksFilename;

	ofstream GFile;
	ofstream DGFile;
	ofstream TFile;
	ofstream GPeaksFile;

	unsigned int errorCount = 0;
	STRINGTONUMBER_STATUS conversionOutcome;

	bool TOptionProvided = false;        // write T to a file
	bool GOptionProvided = false;        // write G to a file
	bool DOptionProvided = false;        // write Delta G to a file
	bool POptionProvided = false;        // write peaks in G to a file
	bool dOptionProvided = false;        // specify delimiter
	bool sOptionProvided = false;        // smoothing parameter
	bool VOptionProvided = false;        // print version info to stdout

	bool TFilenameCheck = false;         // check if the next argument is a filename for T
	bool GFilenameCheck = false;         // check if the next argument is a filename for G
	bool DGFilenameCheck = false;        // check if the next argument is a filename for delta G
	bool GPeaksFilenameCheck = false;    // check if the next argument is a filename for peaks in G
	bool inputFilenameProvided = false;

	if( argc < 2 ) {
		printUsage( argv[0] );
		return 1;
	}

	// Process the arguments

	for(int i = 1; i < argc; i++ ) {
		if( argv[i][0] == '-' && argv[i][2] == '\0') {
			switch( argv[i][1] ) {

				// Output the T statistic to a file. Usage: -T [T_FILENAME]
				case 'T':
					if( TOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one output file for T." << endl;
					}
					else {
						TOptionProvided = true;
						TFilenameCheck = true;
					}
					break;

				// Output the G function to a file. Usage: -G [G_FILENAME]
				case 'G':
					if( GOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one output file for G." << endl;
					}
					else {
						GOptionProvided = true;
						GFilenameCheck = true;
					}
					break;

				// Output the Delta G function to a file. Usage: -D [DG_FILENAME]
				case 'D':
					if( DOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one output file for Delta G." << endl;
					}
					else {
						DOptionProvided = true;
						DGFilenameCheck = true;
					}
					break;

				// Output the locations of the peaks in G to a file. Usage: -P [PEAKS_FILENAME]
				case 'P':
					if( POptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one output file for the peaks in G." << endl;
					}
					else {
						POptionProvided = true;
						GPeaksFilenameCheck = true;
					}
					break;

				// Use SMOOTHING_PARAMETER when detecting the peaks in G
				// (the default is 0.005). Usage: -s SMOOTHING_FACTOR
				case 's':
					if( sOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one smoothing factor.\n" << endl;
					}
					else {
						sOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provided a smoothing factor with the -s option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( smoothingParameter, argv[i+1] );
							i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid smoothing factor: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;
							}
							else if( smoothingParameter < 0.0 ) {
								errorCount++;
								cerr << "Error: The smoothing factor must be positive semi-definite.\n"
								     << "       You provided: " << smoothingParameter << "\n" << endl;
							}
						}
					}
					break;

				// Use DELIMITER to separate fields in the output files
				// (the default is a comma). Usage: -d DELIMITER
				case 'd':
					if( dOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one delimiter.\n" << endl;
					}
					else {
						dOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a delimiter with the -d option.\n" << endl;
						}
						else {
							if( strlen(argv[i+1]) > 1 ) {
								errorCount++;
								cerr << "Error: The delimiter can only be a single character.\n"
								     << "       You provided: " << argv[i+1] << "\n" << endl;
							}
							else {
								delimiter = argv[i+1][0];
								i++;
							}
						}
					}
					break;

				// Print version number
				case 'V':
					if( not VOptionProvided ) {
						printVersion();
						VOptionProvided = true;
					}
					break;

				// Filename that starts with '-' or unknown option
				default:
					if( TFilenameCheck ) {
						TFilename = argv[i];
					}
					else if( GFilenameCheck ) {
						GFilename = argv[i];
					}
					else if( DGFilenameCheck ) {
						DGFilename = argv[i];
					}
					else if( GPeaksFilenameCheck ) {
						GPeaksFilename = argv[i];
					}
					else if( not inputFilenameProvided ) {
						inputFilenameProvided = true;
						inputFilename = argv[i];
					}
					else {
						errorCount++;
						cerr << "Error: Unknown option: " << argv[i] << "\n" << endl;
					}

					TFilenameCheck = false;
					GFilenameCheck = false;
					DGFilenameCheck = false;
					GPeaksFilenameCheck = false;
					break;
			}
		}
		else {
		// The argument had more than 1 character so it's a filename or an unknown option
			if( TFilenameCheck ) {
				TFilename = argv[i];
			}
			else if( GFilenameCheck ) {
				GFilename = argv[i];
			}
			else if( DGFilenameCheck ) {
				DGFilename = argv[i];
			}
			else if( GPeaksFilenameCheck ) {
				GPeaksFilename = argv[i];
			}
			else if( not inputFilenameProvided ) {
				inputFilenameProvided = true;
				inputFilename = argv[i];
			}
			else {
				errorCount++;
				cerr << "Error: Unknown option: " << argv[i] << "\n" << endl;
			}

			TFilenameCheck = false;
			GFilenameCheck = false;
			DGFilenameCheck = false;
			GPeaksFilenameCheck = false;
		}
	}


	// The user only wanted the version number
	if( VOptionProvided && argc == 2 ) {
		return 0;
	}

	// Check if we received all the mandatory arguments

	if( not inputFilenameProvided ) {
		errorCount++;
		cerr << "Error: An input filename is required.\n" << endl;
	}

	// Bail if there were any errors processing the arguments
	if( errorCount ) {
		bail( argv[0], errorCount );
	}

	// Import the sequences

	vector<eventSequence> sequences;
	if( not importSequencesFromFile( inputFilename, sequences ) ) {
		bail( argv[0], ++errorCount );
	}

	// Try to open the output files

	if( TOptionProvided ) {
		if( TFilename.empty() ) {
			TFilename = inputFilename;
			insertBeforeExtension( TFilename, "_T" );
		}

		if( not openOutputFile( TFile, TFilename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	if( GOptionProvided ) {
		if( GFilename.empty() ) {
			GFilename = inputFilename;
			insertBeforeExtension( GFilename, "_G" );
		}

		if( not openOutputFile( GFile, GFilename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	if( DOptionProvided ) {
		if( DGFilename.empty() ) {
			DGFilename = inputFilename;
			insertBeforeExtension( DGFilename, "_DG" );
		}

		if( not openOutputFile( DGFile, DGFilename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	if( POptionProvided ) {
		if( GPeaksFilename.empty() ) {
			GPeaksFilename = inputFilename;
			insertBeforeExtension( GPeaksFilename, "_GPeaks" );
		}

		if( not openOutputFile( GPeaksFile, GPeaksFilename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	// Analyze the data

	for( unsigned int i = 0; i < sequences.size(); i++ ) {
		vector<double> g = G( sequences[i] );
		double t = T( g, smoothingParameter );

		Gee.push_back( g );
		Tee.push_back( t );

		// The following are a bit wasteful since they repeat calculations done
		// while calculating T but I don't think they'll be used often

		if( DOptionProvided ) {
			vector<double> deltag = deltaG( g, smoothingParameter );
			deltaGee.push_back( deltag );
		}

		if( POptionProvided ) {
			vector<unsigned int> peak = detectPeaks( g, smoothingParameter );

			// detectPeaks returns the positions of the peaks in the g vector,
			// which is indexed starting at 0, but what we want to output is the
			// corresponding comparison sites, which are indexed starting at 1.
			for( unsigned int j = 0; j < peak.size(); j++ ) {
				peak[j]++;
			}
			GPeaks.push_back( peak );
		}
	}

	// Output the results

	if( TOptionProvided ) {
		TFile.precision( precision );
		printVector( Tee, TFile, '\n' );
		TFile << "\n";
		TFile.close();
	}
	else {
		cout.precision( precision );
		printVector( Tee, cout, '\n' );
		cout << "\n";
	}

	if( GOptionProvided ) {
		GFile.precision( precision );
		printVector( Gee, GFile, delimiter );
		GFile.close();
	}

	if( DOptionProvided ) {
		DGFile.precision( precision );
		printVector( deltaGee, DGFile, delimiter );
		DGFile.close();
	}

	if( POptionProvided ) {
		printVector( GPeaks, GPeaksFile, delimiter );
		GPeaksFile.close();
	}

	return 0;
}
