/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "eventSequence.hh"

#include <istream>
#include <ostream>
#include <vector>
#include <sstream>
#include <algorithm>  // sort

#include "misc.hh"
#include "io.hh"

/*
** std::ostream& operator<<( std::ostream &outputStream, const eventSequence &sequence );
** 
** Inserts a string representation of sequence into outputStream.
** 
** The string has the following format:
** 
** N:e1,e2,e3,...
** 
** where
**     N is the total number of sites in the sequence
**     e1, e2, e3, ... are the sites of the events
*/
std::ostream& operator<<( std::ostream &outputStream, const eventSequence &sequence ) {
	outputStream << sequence.N << ":";
	printVector( sequence.events, outputStream, ',' );

	return outputStream;
}


/*
** std::istream& operator>>( std::istream &inputStream, eventSequence &sequence );
** 
** Attempts to extract a string representation of sequence from inputStream.
** 
** The string is expected to be in the following format:
** 
** N:e1,e2,e3,...
** 
** where
**     N is the total number of sites in the sequence
**     e1, e2, e3, ... are the sites of the events
*/
std::istream& operator>>( std::istream &inputStream, eventSequence &sequence ) {
	std::string line;
	unsigned int N = 0;  // total number of events
	std::vector<unsigned int> events;
	STRINGTONUMBER_STATUS conversionOutcome;

	if( !getline( inputStream, line ) ) {
		return inputStream;
	}

	// parse our sequence length
	size_t NDelimiterPos = line.find_first_of( ':' );
	if( NDelimiterPos != std::string::npos ) {

		std::string NString = line.substr( 0, NDelimiterPos );
		trimTrailingWhitespace( NString );
		conversionOutcome = stringToNumber( N, NString );

		// valid sequence length
		if( conversionOutcome == SUCCESS ) {

			// parse our events
			std::vector<std::string> tokens = tokanize( line.substr(NDelimiterPos+1), ',' );
			for( unsigned int i = 0; i < tokens.size(); i++ ) {
				unsigned int e;
				trimTrailingWhitespace( tokens[i] );
				conversionOutcome = stringToNumber( e, tokens[i] );

				if( conversionOutcome == SUCCESS ) {
					if( e > 0 && e <= N ) {
						// valid event
						events.push_back( e );
					}
					else {
						// event out of bounds
						std::cerr << "Error: Invalid event: Out of bounds [1, " << N << "]\n"
						          << "       You provided: " << e << "\n" << std::endl;
						inputStream.setstate( std::istream::failbit );
						break;
					}
				}
				else {
					// could not convert to number
					std::cerr << "Error: Invalid event: ";
					printStringToNumberError( conversionOutcome );
					std::cerr << "       You provided: " << tokens[i] << "\n" << std::endl;
					inputStream.setstate( std::istream::failbit );
					break;
				}
			}

			if( inputStream ) {
				// success
				std::sort( events.begin(), events.end() );
				sequence.N = N;
				sequence.events = events;
			}
		}
		else {
			// could not convert to number
			std::cerr << "Error: Invalid sequence length: ";
			printStringToNumberError( conversionOutcome );
			std::cerr << "       You provided: " << NString << "\n" << std::endl;
			inputStream.setstate( std::istream::failbit );
		}
	}
	else {
		// could not find ':'
		std::cerr << "Error: Could not find a sequence length (missing ':')\n"
		          << "       You provided: "
		          << line
		          << "\n" << std::endl;
		inputStream.setstate( std::istream::failbit );
	}

	return inputStream;
}
