/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "io.hh"

#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include "eventSequence.hh"


/*
** bool importSequencesFromFile( const std::string filename, std::vector<eventSequence> &importedSequences );
**
** DESCRIPTION:
**
** Imports sequences from filename. Any imported sequences are appended to the
** vector importedSequences.
**
** If an error occurs, an error message is printed to standard error.
**
** The file is expected to contain one sequence per line with each sequence having
** the following format:
**
** N:e1,e2,e3,...
**
** where
**     N is the total number of sites in the sequence
**     e1, e2, e3, ... are the sites of the events
**
** PARAMETERS:
**     const std::string filename                       the name of the file to import sequences from
**     std::vector<eventSequence> &importedSequences    where to store the imported sequences
**
** RETURNS:
**     true  if sequences were successfully imported from the file
**     false on failure to import sequences
*/
bool importSequencesFromFile( const std::string filename, std::vector<eventSequence> &importedSequences ) {
	std::ifstream inputFile( filename.c_str() );

	if( !inputFile ) {
		std::cerr << "Error: The file " << filename << " could not be opened for input.\n" << std::endl;
		return false;
	}

	size_t originalSize = importedSequences.size();

	while( inputFile ) {
		eventSequence sequence;
		if( inputFile >> sequence ) {
			importedSequences.push_back( sequence );
		}
	}

	inputFile.close();

	// did we actually successfully import any sequences?
	if( importedSequences.size() == originalSize ) {
		std::cerr << "Error: No sequences were found in " << filename << "\n" << std::endl;
		return false;
	}

	return true;
}


/*
** bool openOutputFile( std::ofstream &file, const std::string filename );
**
** DESCRIPTION:
**
** Opens filename for output. Does a simple check to see whether the file
** already exists, refusing to open it if it does.
**
** An error message is printed to standard error if the file fails to open.
**
** IMPORTANT: This test cannot distinguish between a file not existing and
**            one existing that has write, but not read, permission.
**            THIS CAN RESULT IN DATA LOSS.
**
** PARAMETERS:
**     std::ofstream &file     the output stream to associate the file
**     std::string filename    the name of the file to open
**
** RETURNS:
**     true  if the file open successfully
**     false if the file failed to open
*/
bool openOutputFile( std::ofstream &file, const std::string filename ) {
	/*
	** Check if the file already exists by trying to read from it
	**
	** IMPORTANT: This test cannot distinguish between a file not existing and
	**            one existing that has write, but not read, permission.
	**            THIS CAN RESULT IN DATA LOSS.
	*/
	std::ifstream existanceCheck;
	existanceCheck.open( filename.c_str(), std::ios_base::out | std::ios_base::in );
	if( existanceCheck.is_open() ) {
		std::cerr << "Error: Cannot write to file " << filename << " because it already exists.\n" << std::endl;
		return false;
	}
	existanceCheck.close();

	// Go ahead and open it for output
	file.open( filename.c_str() );
	if( !file ) {
		std::cerr << "Error: Could not open " << filename << " for output." << "\n" << std::endl;
		return false;
	}
	else {
		return true;
	}
}


/*
** void insertBeforeExtension( std::string &filename, const std::string insertString );
**
** DESCRIPTION:
**
** Inserts insertString in filename before the last occurrence of '.' (i.e.,
** the file extension). If '.' does not appear in filename, insertString is
** appended.
**
** PARAMETERS:
**     std::string &filename             string to be modified
**     const std::string insertString    string to be inserted
*/
void insertBeforeExtension( std::string &filename, const std::string insertString ) {
	size_t extensionPosition = filename.find_last_of( '.' );

	if( extensionPosition != filename.npos ) {
		filename.insert( extensionPosition, insertString.c_str() );
	}
	else {
		filename.append( insertString );
	}
}


/*
** std::vector<std::string> tokanize( const std::string input, const char delimiter );
**
** DESCRIPTION:
**
** Splits input by delimeter.
**
** PARAMETERS:
**     const std::string input    the string to tokanize
**     const char delimeter       the delimeter
**
** RETURNS:
**     a vector of tokens
*/
std::vector<std::string> tokanize( const std::string input, const char delimiter ) {
	std::stringstream ss(input);
	std::string token;
	std::vector<std::string> tokens;

	while( getline( ss, token, delimiter ) ) {
		tokens.push_back( token );
	}

	return tokens;
}
