/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "misc.hh"

#include <vector>
#include <string>
#include <iostream>
#include <cerrno>     // errno, ERANGE, EINVAL
#include <cmath>      // HUGE_VAL, ceil, floor
#include <cstring>    // strlen
#include <cstdlib>    // strtol
#include <climits>    // LONG_MIN, LONG_MAX, INT_MIN, INT_MAX, ULONG_MAX, UINT_MAX


/*
** void trimTrailingWhitespace( std::string &str );
**
** DESCRIPTION:
**
** Removes trailing whitespace from a std::string.
**
** The characters removed are:
**     ' '     space
**     '\t'    horizontal tab
**     '\v'    vertical tab
**     '\n'    newline
**     '\r'    carriage return
**     '\f'    form feed
**
** PARAMETERS:
**     std::string &str    string to remove trailing whitespace from
*/
void trimTrailingWhitespace( std::string &str ) {
	std::string whitespaces(" \t\v\n\r\f");
	size_t lastNonWhitespace = str.find_last_not_of( whitespaces );

	if( lastNonWhitespace != std::string::npos ) {
		str.erase( lastNonWhitespace+1 );
	}
	else {
		// string is all whitespace
		str.clear();
	}
}


/*
** enum STRINGTONUMBER_STATUS stringToNumber( int &i, const char *str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a C string representation of a number to a signed integer. Ignores
** leading whitespace but does not tolerate any trailing characters including
** whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     int &i                   where to store the result of the conversion
**     const char *str          the string to convert
**     const int base           the base the number should be interpreted in
**                                  default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( int &i, const char *str, const int base ) {
	char *end;    // will point to the first character after the number
	errno = 0;    // reset errno
	long result = std::strtol( str, &end, base );

	if( errno == EINVAL ) {
		return INCONVERTIBLE_ERROR;
	}

	if( (errno == ERANGE && result == LONG_MAX) || result > INT_MAX ) {
		return OVERFLOW_ERROR;
	}

	if( (errno == ERANGE && result == LONG_MIN) || result < INT_MIN ) {
		return UNDERFLOW_ERROR;
	}

	if( *str == '\0' || *end != '\0' ) {
		// original string was empty or there are trailing characters after the number
		return INCONVERTIBLE_ERROR;
	}

	i = (int)result;
	return SUCCESS;
}


/*
** enum STRINGTONUMBER_STATUS stringToNumber( int &i, const std::string str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a std::string representation of a number to a signed integer. Ignores
** leading whitespace but does not tolerate any trailing characters including
** whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     int &i                   where to store the result of the conversion
**     const std::string str    the string to convert
**     const int base           the base the number should be interpreted in
**                                  default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( int &i, const std::string str, const int base ) {
	return stringToNumber( i, str.c_str(), base );
}


/*
** enum STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const char *str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a C string representation of a number to an unsigned integer.
** Ignores leading whitespace but does not tolerate any trailing characters,
** including whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     unsigned int &i          where to store the result of the conversion
**     const char *str          the string to convert
**     const int base           the base the number should be interpreted in
**                                  the default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const char *str, const int base ) {
	char *end;
	errno = 0;
	unsigned long result = std::strtoul( str, &end, base );

	if( errno == EINVAL ) {
		return INCONVERTIBLE_ERROR;
	}

	if( (errno == ERANGE && result == ULONG_MAX) || result > UINT_MAX ) {

		// check if it was an underflow
		errno=0;
		long signedResult = std::strtol( str, &end, base );
		if( errno == 0 && signedResult < 0 ) {
			return UNDERFLOW_ERROR;
		}

		return OVERFLOW_ERROR;
	}

	if( *str == '\0' || *end != '\0' ) {
		return INCONVERTIBLE_ERROR;
	}

	i = (unsigned int)result;
	return SUCCESS;
}


/*
** enum STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const std::string str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a std::string representation of a number to an unsigned integer.
** Ignores leading whitespace but does not tolerate any trailing characters,
** including whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     unsigned int &i          where to store the result of the conversion
**     const std::string str    the string to convert
**     const int base           the base the number should be interpreted in
**                                  the default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const std::string str, const int base ) {
	return stringToNumber( i, str.c_str(), base );
}


/*
** enum STRINGTONUMBER_STATUS stringToNumber( double &d, const char *str );
**
** DESCRIPTION:
**
** Converts a C string representation of a number to a double. Ignores
** leading whitespace but does not tolerate any trailing characters, including
** whitespace.
**
** PARAMETERS:
**     double &d                where to store the result of the conversion
**     const char *str          the string to convert
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( double &d, const char *str ) {
	char *end;
	errno = 0;
	double result = strtod( str, &end );

	if( errno == ERANGE ) {
		if( result == HUGE_VAL ) {
			return OVERFLOW_ERROR;
		}
		else {
			return UNDERFLOW_ERROR;
		}
	}

	if( *str == 0 || *end != '\0' ) {
		return INCONVERTIBLE_ERROR;
	}

	d = result;
	return SUCCESS;
}


/*
** enum STRINGTONUMBER_STATUS stringToNumber( double &d, const std::string str );
**
** DESCRIPTION:
**
** Converts a std::string representation of a number to a double. Ignores
** leading whitespace but does not tolerate any trailing characters, including
** whitespace.
**
** PARAMETERS:
**     double &d                where to store the result of the conversion
**     const std::string str    the string to convert
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( double &d, const std::string str ) {
	return stringToNumber( d, str.c_str() );
}


/*
** void printStringToNumberError( enum STRINGTONUMBER_STATUS status );
**
** DESCRIPTION:
**
** Prints an error message to standard error associated with an
** enum STRINGTONUMBER_STATUS status code.
**
** PARAMETERS:
**     enum STRINGTONUMBER_STATUS status    the status code
*/
void printStringToNumberError( enum STRINGTONUMBER_STATUS status ) {
	switch( status ) {
		case OVERFLOW_ERROR:
			std::cerr << "The number was too large." << std::endl;
			break;

		case UNDERFLOW_ERROR:
			std::cerr << "The number was too small." << std::endl;
			break;

		case INCONVERTIBLE_ERROR:
			std::cerr << "Could not convert to a number." << std::endl;
			break;

		case SUCCESS:
			break;
	}
}
