/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef MISC_HH
#define MISC_HH

#include <vector>
#include <string>


/*
** void trimTrailingWhitespace( std::string &str );
**
** DESCRIPTION:
**
** Removes trailing whitespace from a std::string.
**
** The characters removed are:
**     ' '     space
**     '\t'    horizontal tab
**     '\v'    vertical tab
**     '\n'    newline
**     '\r'    carriage return
**     '\f'    form feed
**
** PARAMETERS:
**     std::string &str    string to remove trailing whitespace from
*/
void trimTrailingWhitespace( std::string &str );


// status codes for the various stringToNumber functions
enum STRINGTONUMBER_STATUS { SUCCESS, OVERFLOW_ERROR, UNDERFLOW_ERROR, INCONVERTIBLE_ERROR };


/*
** enum STRINGTONUMBER_STATUS stringToNumber( int &i, const char *str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a C string representation of a number to a signed integer. Ignores
** leading whitespace but does not tolerate any trailing characters including
** whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     int &i                   where to store the result of the conversion
**     const char *str          the string to convert
**     const int base           the base the number should be interpreted in
**                                  default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( int &i, const char *str, const int base=0 );


/*
** enum STRINGTONUMBER_STATUS stringToNumber( int &i, const std::string str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a std::string representation of a number to a signed integer. Ignores
** leading whitespace but does not tolerate any trailing characters including
** whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     int &i                   where to store the result of the conversion
**     const std::string str    the string to convert
**     const int base           the base the number should be interpreted in
**                                  default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( int &i, const std::string str, const int base=0 );


/*
** enum STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const char *str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a C string representation of a number to an unsigned integer.
** Ignores leading whitespace but does not tolerate any trailing characters,
** including whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     unsigned int &i          where to store the result of the conversion
**     const char *str          the string to convert
**     const int base           the base the number should be interpreted in
**                                  the default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const char *str, const int base=0 );


/*
** enum STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const std::string str, const int base=0 );
**
** DESCRIPTION:
**
** Converts a std::string representation of a number to an unsigned integer.
** Ignores leading whitespace but does not tolerate any trailing characters,
** including whitespace.
**
** Based on: http://stackoverflow.com/questions/194465/how-to-parse-a-string-to-an-int-in-c
**
** PARAMETERS:
**     unsigned int &i          where to store the result of the conversion
**     const std::string str    the string to convert
**     const int base           the base the number should be interpreted in
**                                  the default value is 0, which attempts to determine the base
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
STRINGTONUMBER_STATUS stringToNumber( unsigned int &i, const std::string str, const int base=0 );


/*
** enum STRINGTONUMBER_STATUS stringToNumber( double &d, const char *str );
**
** DESCRIPTION:
**
** Converts a C string representation of a number to a double. Ignores
** leading whitespace but does not tolerate any trailing characters, including
** whitespace.
**
** PARAMETERS:
**     double &d                where to store the result of the conversion
**     const char *str          the string to convert
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( double &d, const char *str );


/*
** enum STRINGTONUMBER_STATUS stringToNumber( double &d, const std::string str );
**
** DESCRIPTION:
**
** Converts a std::string representation of a number to a double. Ignores
** leading whitespace but does not tolerate any trailing characters, including
** whitespace.
**
** PARAMETERS:
**     double &d                where to store the result of the conversion
**     const std::string str    the string to convert
**
** RETURNS:
**     A status code from enum STRINGTONUMBER_STATUS which can be:
**         SUCCESS                no errors occurred
**         INCONVERTIBLE_ERROR    conversion failed
**         OVERFLOW_ERROR         integer overflow occurred
**         UNDERFLOW_ERROR        integer underflow occurred
*/
enum STRINGTONUMBER_STATUS stringToNumber( double &d, const std::string str );


/*
** void printStringToNumberError( enum STRINGTONUMBER_STATUS error );
**
** DESCRIPTION:
**
** Prints an error message to standard error associated with an
** enum STRINGTONUMBER_STATUS status code.
**
** PARAMETERS:
**     enum STRINGTONUMBER_STATUS error    the status code
*/
void printStringToNumberError( enum STRINGTONUMBER_STATUS error );

/*******************************************************************************
**  Templates                                                                 **
*******************************************************************************/

/*
** template<class T>
** inline int findMinimum( const std::vector<T> &data );
**
** DESCRIPTION:
**
** Returns the index of the first occurrence of the minimum value in the vector data.
**
** PARAMETERS:
**     const std::vector<T> &data    the vector of data whose minimum we want
**                                   to find
**
** RETURNS:
**    the index of the first occurrence of the minimum value
**    -1 if data is an empty vector
*/
template<class T>
inline int findMinimum( const std::vector<T> &data );


/*
** template<class T>
** inline int findMaximum( const std::vector<T> &data );
**
** DESCRIPTION:
**
** Returns the index of the first occurrence of the maximum value in the vector data.
**
** PARAMETERS:
**     const std::vector<T> &data    the vector of data whose maximum we want
**                                   to find
**
** RETURNS:
**    the index of the first occurrence of the maximum value
**    -1 if data is an empty vector
*/
template<class T>
inline int findMaximum( const std::vector<T> &data );


/*
** template<class T>
** inline std::vector<T> abs( const std::vector<T> &x );
**
** DESCRIPTION:
**
** Computes the absolute value of each element in the vector x. This creates a new
** vector leaving the original vector is not modified.
**
** PARAMETERS:
**     const std::vector<T> &x    what to compute the absolute value of
**
** RETURNS:
**     a vector containing the absolute values of x
*/
template<class T>
inline std::vector<T> abs( const std::vector<T> &x );


/*
** template<class T>
** std::vector<T> select( const std::vector<unsigned int> &indices, const std::vector<T> &data );
**
** DESCRIPTION:
**
** Picks out the elements of the vector data at the locations given in the
** vector indices. Out of bounds indices are ignored.
**
** PARAMETERS:
**     const std::vector<unsigned int> &indices    indices of the elements we want to select
**     const std::vector<T> &data                  the data we want to select from
**
** RETURNS:
** 	std::vector<T>    vector of values at the desired indices
*/
template<class T>
inline std::vector<T> select( const std::vector<unsigned int> &indices, const std::vector<T> &data );


/*
** template<class T>
** inline bool binarySearch( const std::vector<T> &v, const T x, unsigned int &position );
**
** DESCRIPTION:
**
** Performs a binary search on std::vector<T> &v that is sorted in ascending order.
**
** PARAMETERS:
**     const std::vector<T> &v    what we're looking in
**     const T x                  what we're looking for
**     unsigned int &position     position of x if found otherwise position to insert x
**
** ASSUMPTIONS:
**     const std::vector<T> must be sorted in ascending order
**
** RETURNS:
**     true    If x is in the vector v. The index of x is stored in position.
**     false   If x is not in the vector. The index where x should be inserted is stored in position.
*/
template<class T>
inline unsigned int binarySearch( const std::vector<T> &v, const T x );


// template implementations
#include "misc_templates.cpp"

#endif  // MISC_HH
