/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef MISC_TEMPLATES_CPP
#define MISC_TEMPLATES_CPP

#include <vector>
#include <iostream>
#include <algorithm>  // swap
#include <cmath>      // abs

#include "misc.hh"


/*
** template<class T>
** inline int findMinimum( const std::vector<T> &data );
**
** DESCRIPTION:
**
** Returns the index of the first occurrence of the minimum value in the vector data.
**
** PARAMETERS:
**     const std::vector<T> &data    the vector of data whose minimum we want
**                                   to find
**
** RETURNS:
**    the index of the first occurrence of the minimum value
**    -1 if data is an empty vector
*/
template<class T>
inline int findMinimum( const std::vector<T> &data ) {
	if( data.empty() ) {
		return -1;
	}

	int minimum = 0;

	for( unsigned int i = 1; i < data.size(); i++ ) {
		if( data[i] < data[minimum] ) {
			minimum = i;
		}
	}

	return minimum;
}


/*
** template<class T>
** inline int findMaximum( const std::vector<T> &data );
**
** DESCRIPTION:
**
** Returns the index of the first occurrence of the maximum value in the vector data.
**
** PARAMETERS:
**     const std::vector<T> &data    the vector of data whose maximum we want
**                                   to find
**
** RETURNS:
**    the index of the first occurrence of the maximum value
**    -1 if data is an empty vector
*/
template<class T>
inline int findMaximum( const std::vector<T> &data ) {
	if( data.empty() ) {
		return -1;
	}

	int maximum = 0;

	for( unsigned int i = 1; i < data.size(); i++ ) {
		if( data[i] > data[maximum] ) {
			maximum = i;
		}
	}

	return maximum;
}


/*
** template<class T>
** inline std::vector<T> abs( const std::vector<T> &x );
**
** DESCRIPTION:
**
** Computes the absolute value of each element in the vector x. This creates a new
** vector leaving the original vector is not modified.
**
** PARAMETERS:
**     const std::vector<T> &x    what to compute the absolute value of
**
** RETURNS:
**     a vector containing the absolute values of x
*/
template<class T>
inline std::vector<T> abs( const std::vector<T> &x ) {
	std::vector<T> absx;

	for( unsigned int i = 0; i < x.size(); i++ ) {
		absx.push_back( std::abs(x[i]) );
	}

	return absx;
}


/*
** template<class T>
** std::vector<T> select( const std::vector<unsigned int> &indices, const std::vector<T> &data );
**
** DESCRIPTION:
**
** Picks out the elements of the vector data at the locations given in the
** vector indices. Out of bounds indices are ignored.
**
** PARAMETERS:
**     const std::vector<unsigned int> &indices    indices of the elements we want to select
**     const std::vector<T> &data                  the data we want to select from
**
** RETURNS:
** 	std::vector<T>    vector of values at the desired indices
*/
template<class T>
inline std::vector<T> select( const std::vector<unsigned int> &indices, const std::vector<T> &data ) {
	std::vector <T> selectedData;

	for( unsigned int i = 0; i < indices.size(); i++ ) {
		if( indices[i] < data.size() ) {
			selectedData.push_back( data[indices[i]] );
		}
		else {
			// ignore out of bounds indices
		}
	}

	return selectedData;
}


/*
** template<class T>
** inline bool binarySearch( const std::vector<T> &v, const T x, unsigned int &position );
**
** DESCRIPTION:
**
** Performs a binary search on std::vector<T> &v that is sorted in ascending order.
**
** PARAMETERS:
**     const std::vector<T> &v    what we're looking in
**     const T x                  what we're looking for
**     unsigned int &position     position of x if found otherwise position to insert x
**
** ASSUMPTIONS:
**     const std::vector<T> must be sorted in ascending order
**
** RETURNS:
**     true    If x is in the vector v. The index of x is stored in position.
**     false   If x is not in the vector. The index where x should be inserted is stored in position.
*/
template<class T>
inline bool binarySearch( const std::vector<T> &v, const T x, unsigned int &position ) {
	if( v.empty() ) {
		// Not found and should insert at the beginning
		position = 0;
		return false;
	}

	unsigned int start = 0;
	unsigned int end = v.size()-1;    // no underflow because v.size() >= 1 (list is not empty)
	unsigned int mid = (start+end)/2;

	while( mid != start ) {  // sufficient to check mid != start because mid == start implies start == end
		if( v[mid] == x ) {
		    // Found at position mid
			position = mid;
			return true;
		}
		else if( x < v[mid] ) {
			end = mid;
		}
		else if( x > v[mid] ) {
			start = mid;
		}

		mid = (start+end)/2;
	}


	if( v[start] == x ) {
		// Found at position start
		position = start;
		return true;
	}
	else if( v[end] == x ) {
		// Found at position end
		position = end;
		return true;
	}
	else if( x < v[start] ) {
		// Not found and should insert at the beginning of the vector
		position = start;
		return false;
	}
	else if( x > v[end] ) {
		// Not found and should insert at the end of the vector
		position = end+1;
		return false;
	}
	else {
		// Not found and should insert at position end
		position = end;
		return false;
	}
}

#endif  // MISC_TEMPLATES_CPP
