/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "random.hh"

#include <cstdlib>    // seed48, drand48, strtol, strtod
#include <ctime>      // time
#include <vector>

#include "misc.hh"

extern "C" double drand48( void );
extern "C" unsigned short * seed48( unsigned short xseed[3] );


/*
** void seed_drand48( void ) {
**
** DESCRIPTION:
**
** Seeds drand48 using random numbers generated by rand, which is seeded by the
** current time.
*/
void seed_drand48( void ) {
	srand(time(NULL));
	unsigned short seed[3];

	seed[0] = (unsigned short)rand();
	seed[1] = (unsigned short)rand();
	seed[2] = (unsigned short)rand();

	seed48(seed);
}


/*
** unsigned int randomUnsignedInt( const unsigned int lowLimit, const unsigned int highLimit );
**
** DESCRIPTION:
**
** Returns a random unsigned integer in the interval [lowLimit, highLimit].
**
** The random numbers are generated using drand48 to be consistent with:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const unsigned int lowLimit     lowest random number possible
**     const unsigned int highLimit    highest random number possible
**
** RETURNS:
**     A random unsigned integer in the interval [lowLimit, highLimit].
*/
unsigned int randomUnsignedInt( const unsigned int lowLimit, const unsigned int highLimit ) {
	/*
	** The drand48 function returns a double chosen from a uniform probability
	** distribution over the interval [0.0, 1.0).
	**
	** Note that Version 2 of The Single UNIX Specification (1997) incorrectly
	** states that the interval is [0.0, 1.0]:
	**
	** http://pubs.opengroup.org/onlinepubs/007908799/xsh/drand48.html
	**
	** This was corrected in later versions of the standard such as:
	**
	** The Open Group Base Specifications, Issue 6, 2004:
	** http://pubs.opengroup.org/onlinepubs/009695399/functions/drand48.html
	*/

	if( lowLimit < highLimit ) {
		return (unsigned int)(drand48() * (highLimit-lowLimit+1)) + lowLimit;
	}
	else {
		return (unsigned int)(drand48() * (lowLimit-highLimit+1)) + highLimit;
	}
}


/*
** std::vector<unsigned int> generateUniqueRandomUnsignedInts( const unsigned int a, const unsigned int b, const unsigned int n );
**
** DESCRIPTION:
**
** Returns a vector containing n unique randomly chosen numbers from the interval [a, b].
**
** If n > |b - a| + 1 (i.e., you asked for more numbers than are possible), the
** vector will only have |b - a| + 1 elements.
**
** PARAMETERS:
**     const unsigned int a    lowest possible random number
**     const unsigned int b    highest possible random number
**     const unsigned int n    the number of unique random numbers to generate
**
**
** RETURNS:
**     A vector containing n unique randomly chosen numbers from the interval [a, b].
*/
std::vector<unsigned int> generateUniqueRandomUnsignedInts( const unsigned int a, const unsigned int b, const unsigned int n ) {
	std::vector<unsigned int> randomNumbers;
	unsigned int totalNumbers = 0;

	if( n < 1 ) {
		return randomNumbers;
	}

	unsigned int lowLimit = a;
	unsigned int highLimit = b;

	if( lowLimit > highLimit ) {
		lowLimit = b;
		highLimit = a;
	}

	if( n <= 0.5*(highLimit-lowLimit+1) ) {
	// Generate random numbers to keep
		randomNumbers.push_back( randomUnsignedInt(lowLimit, highLimit) );
		totalNumbers++;

		while( totalNumbers < n ) {
			unsigned int randomNumber = randomUnsignedInt( lowLimit, highLimit );
			unsigned int position = 0;

			if( !binarySearch( randomNumbers, randomNumber, position ) ) {
				// generated a unique number
				randomNumbers.insert( randomNumbers.begin()+position, randomNumber );
				totalNumbers++;
			}
		}
	}
	else {
	// Instead of generating the numbers to keep, generate the numbers to remove
		for( unsigned int i = lowLimit; i <= highLimit; i++ ) {
			randomNumbers.push_back(i);
			totalNumbers++;
		}

		// The number of numbers requested is more than the number of unique numbers
		if( n >= (highLimit-lowLimit+1) ) {
			return randomNumbers;
		}

		while( totalNumbers > n ) {
			unsigned int randomNumber = randomUnsignedInt( lowLimit, highLimit );
			unsigned int position = 0;

			// found an existing number
			if( binarySearch( randomNumbers, randomNumber, position ) ) {
				randomNumbers.erase( randomNumbers.begin()+position );
				totalNumbers--;
			}
		}
	}

	return randomNumbers;
}
