/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


/*
** PROGRAM: randomSequence
** AUTHOR: David C. M. Ostapchuk
**
** DESCRIPTION:
** 
** This program generates comparison sequences of a given length with randomly
** located variant sites. This is used to estimate the probability distribution of
** the T statistic under the null hypothesis in the statistical method
** described in the paper:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
*/


#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>  // sort
#include <cstdlib>    // exit

#include "eventSequence.hh"
#include "random.hh"
#include "io.hh"
#include "misc.hh"

using namespace std;

const string version = "1.2";

void printVersion(void) {
	cout << "randomSequence " << version << endl;
}


vector<eventSequence> generateRandomSequences( const unsigned int numberOfEvents, const unsigned int sequenceLength, const unsigned int numberOfSequences ) {
	seed_drand48();

	vector<eventSequence> randomSequences;

	for( unsigned int i = 0; i < numberOfSequences; i++ ) {
		eventSequence sequence;
		sequence.events = generateUniqueRandomUnsignedInts( 1, sequenceLength, numberOfEvents );
		sort( sequence.events.begin(), sequence.events.end() );
		sequence.N = sequenceLength;
		randomSequences.push_back( sequence );
	}

	return randomSequences;
}


void printUsage( const char *executableName ) {
	cout << "Usage:\n"
         << executableName
	     << " -n NUMBER_OF_EVENTS"
	     << " -N SEQUENCE_LENGTH"
	     << " [OPTIONS]\n\n"
		 << "Generates random SEQUENCE_LENGTH site comparison sequences\n"
	     << "containing NUMBER_OF_EVENTS events each.\n\n"
	     << "Options:\n"
	     << "    -r NUMBER_OF_SEQUENCES     Number of sequences to generate\n"
	     << "    -o OUTPUT_FILENAME         Writes sequences to OUTPUT_FILENAME\n"
	     << "                               instead of standard output\n"
	     << "    -V                         Print the version number"
	     << endl;
}


void bail( const char *executableName, const unsigned int errorCount ) {
	printUsage( executableName );
	exit( errorCount );
}


int main( int argc, char *argv[] ) {
	unsigned int numberOfSequences = 1;
	unsigned int numberOfEvents;
	unsigned int sequenceLength;
	char *filename = NULL;

	unsigned int errorCount = 0;
	bool nOptionProvided = false;  // number of events
	bool NOptionProvided = false;  // number of sites
	bool rOptionProvided = false;  // number of sequences to generate
	bool oOptionProvided = false;  // output sequences to a file
	bool VOptionProvided = false;  // print version info to stdout
	STRINGTONUMBER_STATUS conversionOutcome;

	if( argc < 2 ) {
		printUsage( argv[0] );
		return 0;
	}

	// Process the arguments

	for( int i = 1; i < argc; i++ ) {
		if( argv[i][0] == '-' && argv[i][2] == '\0') {
			switch( argv[i][1] ) {

				// Number of events per sequence. Usage: -n NUMBER_OF_EVENTS
				case 'n':
					if( nOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one number of events per sequence.\n" << endl;
					}
					else {
						nOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a events per sequence with the -n option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( numberOfEvents, argv[i+1] );
							i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid number of events: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;
							}
						}
					}
					break;

				// Sequence length (number of comparison sites). Usage: -N SEQUENCE_LENGTH
				case 'N':
					if( NOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one sequence length.\n" << endl;
					}
					else {
						NOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a sequence length with the -N option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( sequenceLength, argv[i+1] );
							i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid sequence length: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;
							}
							else if( sequenceLength < 1 ) {
								errorCount++;
								cerr << "Error: The sequence length must be at least 1.\n"
								     << "       You provided: " << sequenceLength << "\n" << endl;
							}
						}
					}
					break;

				// Number of sequences to generate. Usage: -r NUMBER_OF_SEQUENCES
				case 'r':
					if( rOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one total number of sequences.\n" << endl;
					}
					else {
						rOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a total number of sequences with the -r option\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( numberOfSequences, argv[i+1] );
						    i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid number of sequences: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;;
							}
							else if( numberOfSequences < 1 ) {
								errorCount++;
								cerr << "Error: The total number of sequences must be at least 1.\n"
								     << "       You provided: " << numberOfSequences << "\n" << endl;
							}
						}
					}
					break;

				// Writes sequences to OUTPUT_FILENAME instead of stdout. Usage: -o OUTPUT_FILENAME
				case 'o':
					if( oOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one sequence output file.\n" << endl;
					}
					else {
						oOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a sequence output filename with the -o option.\n" << endl;
						}
						else {
							filename = argv[i+1];
							i++;
						}
					}
					break;

				// Print version number to stderr
				case 'V':
					if( not VOptionProvided ) {
						printVersion();
						VOptionProvided = true;
					}
					break;

				default: // unknown option
					errorCount++;
					cerr << "Error: Unknown option: " << argv[i] << "\n" << endl;
			}
		}
		else {  // option had more than 1 character
			errorCount++;
			cerr << "Error: Unknown option: " << argv[i] << "\n" << endl;
		}
	}


	// The user only wanted the version number
	if( VOptionProvided && argc == 2 ) {
		return 0;
	}

	// Check if we received all the mandatory arguments

	if( not nOptionProvided ) {
		errorCount++;
		cerr << "Error: The number of events per sequence (-n option) is required.\n" << endl;
	}

	if( not NOptionProvided ) {
		errorCount++;
		cerr << "Error: The sequence length (-N option) is required.\n" << endl;
	}

	// Check if the arguments are valid

	if( numberOfEvents > sequenceLength ) {
		errorCount++;
		cerr << "Error: The number of events per sequence must be less than or equal to the sequence length.\n"
		     << "       You provided:\n"
		     << "           n = " << numberOfEvents << "\n"
             << "           N = " << sequenceLength << "\n" << endl;
	}

	// Bail if there were errors processing the arguments
	if( errorCount > 0 ) {
		bail( argv[0], errorCount );
	}

	// Try to open the output file
	ofstream outputFile;
	if( oOptionProvided ) {
		if( not openOutputFile( outputFile, filename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	vector<eventSequence> randomSequences = generateRandomSequences( numberOfEvents, sequenceLength, numberOfSequences );

	if( oOptionProvided ) {
		printVector( randomSequences, outputFile, '\n' );
		outputFile << "\n";
		outputFile.close();
	}
	else {
		printVector( randomSequences, cout, '\n' );
		cout << "\n";
	}

	return 0;
}
