/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "statistics.hh"

#include <vector>
#include <iostream>
#include <cmath>    // fabs, sqrt
#include <cstdlib>  // exit

#include "io.hh"
#include "misc.hh"

#define ZERO_TOLERANCE 1e-16

/*
** std::vector<unsigned int> detectPeaks( std::vector<double> x, double tolerance );
**
** Detects local extrema in the vector x. Changes smaller than tolerance are
** ignored, which is the smoothing technique described in the bottom left
** corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     std::vector<double> x    the data to find the extrema in
**     double tolerance         the minimum change required to count as an extremum
**
** RETURNS:
**     a vector containing the indices of the local extrema in x
*/
std::vector<unsigned int> detectPeaks( std::vector<double> x, double tolerance ) {
	// state variables:
	enum { INCREASING, DECREASING } state;  // an extremum is a change between INCREASING and DECREASING
	bool isBelowTolerance = false;  // to handle extrema that occur in a "flat" region (changes smaller than tolerance)

	std::vector<unsigned int> extrema;         // the detected peaks
	std::vector<unsigned int> belowTolerance;  // temporary storage for "flat" regions (in case a peak appears there)

	if( tolerance < 0.0 ) {
		tolerance = std::fabs(tolerance);
	}

	if( x.empty() ) {
		// no points, no extrema!
		return extrema;
	}

	// the first local extremum is the start point
	extrema.push_back( 0 );

	if( x.size() == 1 ) {
		// if we only have one point, we're done
		return extrema;
	}

	unsigned int i = 1;  // indexes x

	// find first significant change
	while( i < x.size()-1 && std::fabs(x[i] - x[i-1]) < tolerance ) i++;
	if(x[i] - x[i-1] > 0.0 ) {
		state = INCREASING;
	}
	else {
		state = DECREASING;
	}

	for( ; i < x.size(); i++ ) {
		if( x[i] - x[i-1] > tolerance ) {
			// the function is increasing
			if( state == DECREASING ) {
				// local minimum found
				if( isBelowTolerance ) {
					// the minimum is in a "flat" region (changes smaller than tolerance)
					// so choose the smallest value in the region
					belowTolerance.push_back( i-1 );
					int min = findMinimum( select( belowTolerance, x ) );
					extrema.push_back( belowTolerance[min] );
					belowTolerance.clear();
					isBelowTolerance = false;
				}
				else {
					// "sharp" min (the change on both sides of the peak was larger than tolerance)
					extrema.push_back( i-1 );
				}
			}
			else {
				// still increasing
				if( isBelowTolerance ) {
					// there wasn't an extremum in a "flat" region (changes smaller than tolerance)
					belowTolerance.clear();
					isBelowTolerance = false;
				}
			}
			state = INCREASING;

		} else if( x[i] - x[i-1] < -tolerance ) {
			// the function is decreasing
			if( state == INCREASING ) {
	            // local maximum found
				if( isBelowTolerance ) {
					// the maximum is in a "flat" region (changes smaller than tolerance)
					// so choose the largest value in the region
					belowTolerance.push_back( i-1 );
					int max = findMaximum( select( belowTolerance, x ) );
					extrema.push_back( belowTolerance[max] );
					belowTolerance.clear();
					isBelowTolerance = false;
				}
				else {
					// "sharp" max (the change on both sides of the peak was larger than tolerance)
					extrema.push_back( i-1 );
				}
			}
			else {
				if( isBelowTolerance ) {
					// there wasn't an extremum a "flat" region (changes smaller than tolerance)
					belowTolerance.clear();
					isBelowTolerance = false;
				}
			}
			state = DECREASING;

		} else {
			// the change is smaller than our tolerance
			// we'll remember this point in case an extrema occurs here
			belowTolerance.push_back(i-1);
			isBelowTolerance = true;
		}
	}

	// the last local extremum is the end point
	extrema.push_back( x.size()-1 );

	return extrema;
}


/*
** std::vector<double> G( const eventSequence sequence );
**
** DESCRIPTION:
**
** Computes G(x_k), the cumulative difference between the observed and expected
** proportion of events up to position x_k, which is defined as
**
** G(x_k) = F_n(x_k) - F(x_k) = k / n - x_k / N,
**
** where:
**     x_k      is the position of the kth event;
**     F_n(x_k) is the empirical cumulative distribution function (ECDF);
**     F(x_k)   is the cumulative distribution function (CDF);
**     n        is the total number of events in the sequence comparison;
**     N        is the total number of sites in the comparison.
**
** The ECDF is denoted F_n(x_k) and is defined as
** F_n(x_k) = (# of events occurring up to position x_k) / (total number of events)
**          = k / n
**
** The CDF is denoted by F(x_k) and is defined as
** F(x_k) = (position of the kth event) / (total sequence length)
**        = x_k / N
**
** For more information see the top left corner of page 487 of:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const eventSequence sequence    the comparison of two DNA sequences
**
** RETURNS:
**     A vector containing the function G(x_k) computed at all events
*/
std::vector<double> G( const eventSequence sequence ) {
	std::vector<double> g;
	unsigned int n = sequence.events.size();

	for( unsigned int k = 0; k < n; k++ ) {
		g.push_back( (double)(k+1)/n - (double)sequence.events[k]/sequence.N );
	}

	return g;
}


/*
** std::vector<double> deltaG( const std::vector<double> &g, double smoothingParameter );
**
** DESCRIPTION:
**
** Computes \Delta G_{ij}, the change in G(x_k) bound by events i and j, for
** all intervals where G(x_k) is monotonically increasing or decreasing
** (ignored changes smaller than smoothingParameter).
**
** See the bottom left corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector<double> &g    the function G(x_k) evaluated at all events
**     double smoothingParameter       the minimum amount of change required
**                                     for a change in monotonicity
**
** RETURNS:
**     a vector of \Delta G_{ij} for all intervals where G(x_k) is
**     monotonically increasing or decreasing
*/
std::vector<double> deltaG( const std::vector<double> &g, double smoothingParameter ) {
	std::vector<unsigned int> gPeaks = detectPeaks( g, smoothingParameter );

	std::vector<double> dg;

	for( unsigned int i = 1; i < gPeaks.size(); i++ ) {
		dg.push_back( g[gPeaks[i]] - g[gPeaks[i-1]] );
	}

	return dg;
}


/*
** double T( const std::vector<double> &g, double smoothingParameter );
**
** DESCRIPTION:
**
** Computes the test statistic T defined as
**     T = sign( \Delta G ) * max( |\Delta G| )
**
** The maximum is taken over all regions where G is monotonically increasing or
** decreasing (ignoring changes smaller than smoothingParameter).
**
** See the bottom left corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector<double> &g    the function G(x_k) evaluated at all events
**     double smoothingParameter       the minimum change required for a change in monotonicity
**
** RETURNS:
**     The test statistic T
*/
double T( const std::vector<double> &g, double smoothingParameter ) {
	std::vector<double> dg = deltaG( g, smoothingParameter );
	std::vector<double> absdg = abs( dg );
	return dg[findMaximum(absdg)];
}


/*
** std::vector< std::vector<double> > THistogram( const std::vector<double> &data, const unsigned int numberOfBins );
** 
** DESCRIPTION:
** 
** Bins the contents of data into numberOfBins equally sized bins.
** 
** ASSUMPTIONS:
**     the minimum value in data is negative
**     the maximum value in data is positive
** 
** PARAMETERS:
**     const std::vector<double> &data    the data to be binned
**     const unsigned int numberOfBins    the total number of bins to use
** 
** RETURNS:
**     A vector representing the histogram. Each bin is represented by a vector
**     containing two elements:
**         the first element is the lower limit of the bin
**         the second element is the number of counts in the bin
*/
std::vector< std::vector<double> > THistogram( const std::vector<double> &data, const unsigned int numberOfBins ) {
	std::vector< std::vector<double> > h( numberOfBins );
	double min = data[findMinimum(data)];
	double max = data[findMaximum(data)];

	// choose our bins symmetrically about zero
	if( std::fabs(min) > std::fabs(max) )
		max = -min;
	else
		min = -max;

	/*
	** The -2 in the denominator comes from:
	**     -1 because of fenceposting
	**     -1 to shift the bins such that the minimum and maximum values end up
    **        in the middle of the end bins
	*/
	double binSize = std::fabs(max-min)/(numberOfBins-2);

	for( unsigned int i = 0; i < numberOfBins; i++ ) {
		h[i].push_back( min + (i-0.5)*binSize );  // lower limit of the bin
		h[i].push_back( 0 );                      // bin count
	}

	for( unsigned int i = 0; i < data.size(); i++ ) {
		// bins have an interval of (lower, upper]
		int binNumber = (int)( (data[i]+std::fabs(min))/binSize );  // adding std::fabs(min) makes the numerator positive
		h[binNumber][1] += 1;
	}

	return h;
}


/*
** double trapezoidRule( std::vector<double> a, std::vector<double> b);
**
** DESCRIPTION:
**
** Computes the area of the trapezoid formed by the x-axis and the points a and
** b. This is used for numerical integration (trapezoid rule).
**
** PARAMETERS:
**     std::vector<double> a    the first corner of the trapezoid
**     std::vector<double> b    the second corner of the trapezoid
**
**     The first element is the x coordinate, the second is the y coordinate.
**
** RETURNS:
**     The area of the trapezoid
*/
double trapezoidRule( std::vector<double> a, std::vector<double> b) {
	if( a.size() >= 2 || b.size() >= 2 ) {
		return (b[0] - a[0]) * (b[1] + a[1]) / 2.0;
	}
	else {
		return 0.0;
	}
}


/*
** double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples, unsigned int start, unsigned int end );
**
** DESCRIPTION:
**
** Numerically integrates a function using the extended trapezoid rule.
**
** PARAMETERS:
**     std::vector< std::vector<double> > integrandSamples    samples of the integrand
**     unsigned int start                                     the index of the lower integration limit
**     unsigned int end                                       the index of the upper integration limit
**
** RETURNS:
**     the area under the curve
*/
double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples, unsigned int start, unsigned int end ) {
	double area = 0.0;
	int step;
	unsigned int i;

	if( integrandSamples.size() < 2 ) {
		// there is no area under a point
		return 0.0;
	}

	if( start >= integrandSamples.size() ) {
		start = integrandSamples.size()-1;
	}

	if( end >= integrandSamples.size() ) {
		end = integrandSamples.size()-1;
	}

	if( start < end ) {
		step = 1;
	}
	else {
		step = -1;
	}

	i = start;
	while( i != end ) {
		i += step;
		area += trapezoidRule( integrandSamples[i-step], integrandSamples[i] );
	}


	return area;
}


/*
** double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples );
**
** DESCRIPTION:
**
** Numerically integrates a function using the extended trapezoid rule.
**
** PARAMETERS:
**     std::vector< std::vector<double> > integrandSamples    samples of the integrand
**
** RETURNS:
**     the area under the curve
*/
double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples ) {
	return extendedTrapezoidRule( integrandSamples, 0, integrandSamples.size() );
}


/*
** double oneTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel, const bool left );
**
** DESCRIPTION:
**
** Returns the critical value T* for a one tail test at a given rejection level.
**
** To do this it first numerically integrates the probability distribution using
** the extended trapezoid rule. Then, it linearly interpolates between the two
** nearest samples of the probability distribution to determine T*.
**
** For more details see the top right corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector< std::vector<double> > &distribution    samples of the probability distribution
**     const double rejectionLevel                               the rejection level
**     const bool left                                           flag to indicate whether to do a left or right tail test
**
** ASSUMPTIONS:
**     rejectionLevel must be in the interval (0, 1)
**
** RETURNS:
**     T* for the given rejection level for a one tail test
*/
double oneTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel, const bool left ) {
	int step;
	unsigned int begin, end, i;

	if( left ) {
		// left tail test
		begin = 0;
		end = distribution.size()-1;   // last element
		step = 1;
	}
	else {
		// right tail test
		begin = distribution.size()-1; // last element
		end = 0;
		step = -1;
	}

	double totalArea = extendedTrapezoidRule( distribution );
	double cumulativeArea = 0.0;
	double targetArea = rejectionLevel * totalArea;  // assumes rejectionLevel is in the interval (0, 1)

	// integrate until we reach our rejection level
	i = begin;
	while( cumulativeArea < targetArea && i != end ) {
		i += step;
		cumulativeArea += step * trapezoidRule( distribution[i-step], distribution[i] );
	}
	/*
	** i is now the index of the "uncorrected" T* --- it's the sample of the
	** probability distribution at which the area exceeds our rejection level.
	** To get a better estimate of T*, we linearly interpolate between the two
	** nearest samples of the probability distribution.
	*/

	double deltaArea = cumulativeArea - targetArea;  // the excess area (overshoot)

	// handle the case where the slope of the interpolated line is zero
	if( distribution[i][1] - distribution[i-step][1] <  ZERO_TOLERANCE
	 && distribution[i][1] - distribution[i-step][1] > -ZERO_TOLERANCE ) {
		if( distribution[i][1] > ZERO_TOLERANCE || distribution[i][1] < -ZERO_TOLERANCE ) {
			return distribution[i][0] - step*deltaArea/distribution[i][1];
		}
		else {
			// If we've got here the probability distribution is flat and very close to zero, so don't even bother correcting
			return distribution[i][1];
		}
	}

	// slope of the line between the uncorrected T* and the previous point
	double m = (distribution[i][1]-distribution[i-step][1])
	         / (distribution[i][0]-distribution[i-step][0]);

	// possible interpolated T* values
	double alpha = (distribution[i][1]/m - distribution[i][0]);
	double solution1 = -alpha + std::sqrt( (alpha+distribution[i][0])*(alpha+distribution[i][0]) - step*2.0*deltaArea/m );
	double solution2 = -alpha - std::sqrt( (alpha+distribution[i][0])*(alpha+distribution[i][0]) - step*2.0*deltaArea/m );

	// choose the correct solution by checking if it lies between the uncorrected T* and the previous point
	if( distribution[i-step][0] < distribution[i][0] ) {
		if( solution1 >= distribution[i-step][0] && solution1 <= distribution[i][0] ) {
			return solution1;
		}
		else if( solution2 >= distribution[i-step][0] && solution2 <= distribution[i][0] ) {
			return solution2;
		}
		else {
			// should never get here
			std::cerr << "Error: Linear interpolation failed." << std::endl;
			std::exit(1);
		}
	}
	else {
		if( solution1 >= distribution[i][0] && solution1 <= distribution[i-step][0] ) {
			return solution1;
		}
		else if( solution2 >= distribution[i][0] && solution2 <= distribution[i-step][0] ) {
			return solution2;
		}
		else {
			// should never get here
			std::cerr << "Error: Linear interpolation failed." << std::endl;
			std::exit(1);
		}
	}
}


/*
** double twoTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel );
**
** DESCRIPTION:
**
** Returns the critical value T* for a two tail test at a given rejection level.
**
** Since the probability distribution is symmetric, it does this by averaging the
** critical values of a right and a left tail test at half the rejection level.
**
** For more details see the top right corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector< std::vector<double> > &distribution    samples of the probability distribution
**     const double rejectionLevel                               the rejection level
**
** ASSUMPTIONS:
**     rejectionLevel must be in the interval (0, 1)
**
** RETURNS:
**     T* for the given rejection level for a one tail test
*/
double twoTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel ) {
	return 0.5*( std::fabs(oneTailTest( distribution, rejectionLevel/2.0, true ))
	           + std::fabs(oneTailTest( distribution, rejectionLevel/2.0, false )) );
}
