/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef STATISTICS_HH
#define STATISTICS_HH

#include <vector>

#include "eventSequence.hh"


/*
** std::vector<unsigned int> detectPeaks( std::vector<double> x, double tolerance );
**
** Detects local extrema in the vector x. Changes smaller than tolerance are
** ignored, which is the smoothing technique described in the bottom left
** corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     std::vector<double> x    the data to find the extrema in
**     double tolerance         the minimum change required to count as an extremum
**
** RETURNS:
**     a vector containing the indices of the local extrema in x
*/
std::vector<unsigned int> detectPeaks( std::vector<double> x, double tolerance );


/*
** std::vector<double> G( const eventSequence sequence );
**
** DESCRIPTION:
**
** Computes G(x_k), the cumulative difference between the observed and expected
** proportion of events up to position x_k, which is defined as
**
** G(x_k) = F_n(x_k) - F(x_k) = k / n - x_k / N,
**
** where:
**     x_k      is the position of the kth event;
**     F_n(x_k) is the empirical cumulative distribution function (ECDF);
**     F(x_k)   is the cumulative distribution function (CDF);
**     n        is the total number of events in the sequence comparison;
**     N        is the total number of sites in the comparison.
**
** The ECDF is denoted F_n(x_k) and is defined as
** F_n(x_k) = (# of events occurring up to position x_k) / (total number of events)
**          = k / n
**
** The CDF is denoted by F(x_k) and is defined as
** F(x_k) = (position of the kth event) / (total sequence length)
**        = x_k / N
**
** For more information see the top left corner of page 487 of:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const eventSequence sequence    the comparison of two DNA sequences
**
** RETURNS:
**     A vector containing the function G(x_k) computed at all events
*/
std::vector<double> G( const eventSequence sequence );


/*
** std::vector<double> deltaG( const std::vector<double> &g, double smoothingParameter );
**
** DESCRIPTION:
**
** Computes \Delta G_{ij}, the change in G(x_k) bound by events i and j, for
** all intervals where G(x_k) is monotonically increasing or decreasing
** (ignored changes smaller than smoothingParameter).
**
** See the bottom left corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector<double> &g    the function G(x_k) evaluated at all events
**     double smoothingParameter       the minimum amount of change required
**                                     for a change in monotonicity
**
** RETURNS:
**     a vector of \Delta G_{ij} for all intervals where G(x_k) is
**     monotonically increasing or decreasing
*/
std::vector<double> deltaG( const std::vector<double> &g, double smoothingParameter );


/*
** double T( const std::vector<double> &g, double smoothingParameter );
**
** DESCRIPTION:
**
** Computes the test statistic T defined as
**     T = sign( \Delta G ) * max( |\Delta G| )
**
** The maximum is taken over all regions where G is monotonically increasing or
** decreasing (ignoring changes smaller than smoothingParameter).
**
** See the bottom left corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector<double> &g    the function G(x_k) evaluated at all events
**     double smoothingParameter       the minimum change required for a change in monotonicity
**
** RETURNS:
**     The test statistic T
*/
double T( const std::vector<double> &g, double smoothingParameter );


/*
** std::vector< std::vector<double> > THistogram( const std::vector<double> &data, const unsigned int numberOfBins );
** 
** DESCRIPTION:
** 
** Bins the contents of data into numberOfBins equally sized bins.
** 
** ASSUMPTIONS:
**     the minimum value in data is negative
**     the maximum value in data is positive
** 
** PARAMETERS:
**     const std::vector<double> &data    the data to be binned
**     const unsigned int numberOfBins    the total number of bins to use
** 
** RETURNS:
**     A vector representing the histogram. Each bin is represented by a vector
**     containing two elements:
**         the first element is the lower limit of the bin
**         the second element is the number of counts in the bin
*/
std::vector< std::vector<double> > THistogram( const std::vector<double> &data, const unsigned int numberOfBins );


/*
** double trapezoidRule( std::vector<double> a, std::vector<double> b);
**
** DESCRIPTION:
**
** Computes the area of the trapezoid formed by the x-axis and the points a and
** b. This is used for numerical integration (trapezoid rule).
**
** PARAMETERS:
**     std::vector<double> a    the first corner of the trapezoid
**     std::vector<double> b    the second corner of the trapezoid
**
**     The first element is the x coordinate, the second is the y coordinate.
**
** RETURNS:
**     The area of the trapezoid
*/
double trapezoidRule( std::vector<double> a, std::vector<double> b);


/*
** double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples, unsigned int start, unsigned int end );
**
** DESCRIPTION:
**
** Numerically integrates a function using the extended trapezoid rule.
**
** PARAMETERS:
**     std::vector< std::vector<double> > integrandSamples    samples of the integrand
**     unsigned int start                                     the index of the lower integration limit
**     unsigned int end                                       the index of the upper integration limit
**
** RETURNS:
**     the area under the curve
*/
double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples, unsigned int start, unsigned int end );


/*
** double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples );
**
** DESCRIPTION:
**
** Numerically integrates a function using the extended trapezoid rule.
**
** PARAMETERS:
**     std::vector< std::vector<double> > integrandSamples    samples of the integrand
**
** RETURNS:
**     the area under the curve
*/
double extendedTrapezoidRule( std::vector< std::vector<double> > integrandSamples );


/*
** double oneTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel, const bool left=true );
**
** DESCRIPTION:
**
** Returns the critical value T* for a one tail test at a given rejection level.
**
** To do this it first numerically integrates the probability distribution using
** the extended trapezoid rule. Then, it linearly interpolates between the two
** nearest samples of the probability distribution to determine T*.
**
** For more details see the top right corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector< std::vector<double> > &distribution    samples of the probability distribution
**     const double rejectionLevel                               the rejection level
**     const bool left                                           flag to indicate whether to do a left or right tail test
**
** ASSUMPTIONS:
**     rejectionLevel must be in the interval (0, 1)
**
** RETURNS:
**     T* for the given rejection level for a one tail test
*/
double oneTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel, const bool left=true );


/*
** double twoTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel );
**
** DESCRIPTION:
**
** Returns the critical value T* for a two tail test at a given rejection level.
**
** Since the probability distribution is symmetric, it does this by averaging the
** critical values of a right and a left tail test at half the rejection level.
**
** For more details see the top right corner of page 487 in:
**
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** PARAMETERS:
**     const std::vector< std::vector<double> > &distribution    samples of the probability distribution
**     const double rejectionLevel                               the rejection level
**
** ASSUMPTIONS:
**     rejectionLevel must be in the interval (0, 1)
**
** RETURNS:
**     T* for the given rejection level for a one tail test
*/
double twoTailTest( const std::vector< std::vector<double> > &distribution, const double rejectionLevel );

#endif  // STATISTICS_HH
