/*
** Copyright (c) 2012, David C. M. Ostapchuk (d.ostapchuk@uwinnipeg.ca)
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**
** 1. Redistributions of source code must retain the above copyright notice, this
**    list of conditions and the following disclaimer.
**
** 2. Redistributions in binary form must reproduce the above copyright notice,
**    this list of conditions and the following disclaimer in the documentation
**    and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


/*
** PROGRAM: tstar
** AUTHOR: David C. M. Ostapchuk
**
** DESCRIPTION:
** 
** This program computes T*, the critical value of the T statistic, for a given
** sequence length, number of events, and rejection level. This critical value
** is used to determine if variations in DNA sequences are statistically
** significant.
** 
** It does this by estimating the probability distribution function of the T
** statistic by computing it for many randomly generated sequences with a given
** number of events. The probability distribution function is then integrated to
** find the critical value T*.
** 
** Details about this statistical method, including the definition of the T
** statistic can be found in the following paper:
** 
** Tang, H. and Lewonton, R. C. Locating Regions of Differential Variability in
** DNA and Protein Sequences. Genetics 153(1):485-495, September 1999.
** PMID:10471728, PMCID:PMC1460758, WOS:000082421600035.
**
** A selection of precomputed T* values is in table 1 on page 488 of the above
** paper.
*/


#include <vector>
#include <iostream>
#include <fstream>
#include <cstdlib>  // exit
#include <cmath>    // floor
#include <cstring>  // strlen

#include "eventSequence.hh"
#include "random.hh"
#include "io.hh"
#include "misc.hh"
#include "statistics.hh"

using namespace std;

const string version = "1.3";

void printVersion( void ) {
	cout << "tstar " << version << endl;
}


void printUsage( const char *executableName ) {
	cout << "Usage:\n"
	     << executableName
	     << " -i SEQUENCE_FILENAME"
	     << " -c REJECTION_LEVEL"
	     << " [OPTIONS]\n"
         << executableName
	     << " -n NUMBER_OF_EVENTS"
	     << " -N SEQUENCE_LENGTH"
	     << " -r NUMBER_OF_SEQUENCES"
	     << " -c REJECTION_LEVEL"
	     << " [OPTIONS]\n\n"
		 << "Calculates the critical value T* at REJECTION_LEVEL from the comparison\n"
	     << "sequences in SEQUENCE_FILE or NUMBER_OF_SEQUENCES randomly generated\n"
	     << "SEQUENCE_LENGTH site comparison sequences containing NUMBER_OF_EVENTS\n"
	     << "events each. By default it calculates T* for two tail tests and prints it to\n"
	     << "standard output.\n\n"
	     << "Options:\n"
	     << "    -L                        Calculate T* for left tail tests\n"
	     << "    -R                        Calculate T* for right tail tests\n"
	     << "    -T TSTAR_FILENAME         Output T* to TSTAR_FILENAME\n"
	     << "    -o OUTPUT_FILENAME        Output the sequences used to OUTPUT_FILENAME\n"
	     << "                              Cannot be used with the -i option\n"
	     << "    -H HISTOGRAM_FILENAME     Output the T histogram to HISTOGRAM_FILENAME\n"
	     << "    -b BINS                   Use a total of BINS bins when generating the\n"
	     << "                              histogram (the default is sqrt(N*n))\n"
	     << "    -s SMOOTHING_PARAMETER    Use SMOOTHING_PARAMETER when detecting the\n"
	     << "                              peaks in G (the default is 0.005)\n"
	     << "    -d DELIMITER              Use DELIMITER to separate fields in input\n"
	     << "                              and output files (the default is a comma)\n"
	     << "    -V                        Print the version number"
	     << endl;
}


void bail( const char *executableName, const unsigned int errorCount ) {
	printUsage( executableName );
	exit( errorCount );
}


int main( int argc, char *argv[] ) {
	int precision = 10;      // The max precision at which the data is output
	char delimiter = ',';    // default to CSV format
	unsigned int numberOfSequences = 1;
	unsigned int numberOfEvents;
	unsigned int sequenceLength;
	unsigned int numberOfBins;
	double rejectionLevel;
	double smoothingParameter = 0.005;    // default to what was used in Tang and Lewontin, Genetics 153(1):485, 1999

	double tStar;

	vector<eventSequence> sequences;
	vector<double> Ts;
	vector< vector<double> > histogram;

	string sequenceInputFilename;
	string TStarFilename;
	string sequenceOutputFilename;
	string histogramFilename;

	ifstream sequenceInputFile;
	ofstream TStarFile;
	ofstream sequenceOutputFile;
	ofstream histogramFile;

	unsigned int errorCount = 0;
	STRINGTONUMBER_STATUS conversionOutcome;

	bool iOptionProvided = false;
	bool nOptionProvided = false;
	bool NOptionProvided = false;
	bool rOptionProvided = false;
	bool cOptionProvided = false;
	bool LOptionProvided = false;
	bool ROptionProvided = false;
	bool TOptionProvided = false;
	bool oOptionProvided = false;
	bool HOptionProvided = false;
	bool bOptionProvided = false;
	bool sOptionProvided = false;
	bool dOptionProvided = false;
	bool VOptionProvided = false;

	if( argc < 2 ) {
		printUsage( argv[0] );
		return 1;
	}

	// Process the arguments

	for( int i = 1; i < argc; i++ ) {
		if( argv[i][0] == '-' && argv[i][2] == '\0') {
			switch( argv[i][1] ) {

				// Use sequences in SEQUENCE_FILENAME. Usage: -i SEQUENCE_FILENAME
				case 'i':
					if( iOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one sequence input file.\n" << endl;
					}
					else {
						iOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a sequence input filename with the -i option.\n" << endl;
						}
						else {
							sequenceInputFilename = argv[i+1];
							i++;
						}
					}
					break;

				// Number of events per sequence. Usage -n NUMBER_OF_EVENTS
				case 'n':
					if( nOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one number of events per sequence.\n" << endl;
					}
					else {
						nOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a number of events per sequence with the -n option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( numberOfEvents, argv[i+1] );
							i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid number of events per sequence: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;
							}
						}
					}
					break;

				// Sequence length (number of sites). Usage: -N SEQUENCE_LENGTH
				case 'N':
					if( NOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one sequence length.\n" << endl;
					}
					else {
						NOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a sequence length with the -N option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( sequenceLength, argv[i+1] );
							i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid sequence length: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;
							}
							else if( sequenceLength < 1 ) {
								errorCount++;
								cerr << "Error: The sequence length must be at least 1.\n"
								     << "       You provided: " << sequenceLength << "\n" << endl;
							}
						}
					}
					break;

				// Number of sequences to generate. Usage: -r NUMBER_OF_SEQUENCES
				case 'r':
					if( rOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one total number of sequences.\n" << endl;
					}
					else {
						rOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a total number of sequences with the -r option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( numberOfSequences, argv[i+1] );
						    i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid number of sequences: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;;
							}
							else if( numberOfSequences < 1 ) {
								errorCount++;
								cerr << "Error: The total number of sequences must be at least 1.\n"
								     << "       You provided: " << argv[i+1] << "\n" << endl;
							}
						}
					}
					break;

				// Rejection level at which to calculate T*. Usage: -c REJECTION_LEVEL
				case 'c':
					if( cOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one rejection level.\n" << endl;
					}
					else {
						cOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a rejection level with the -c option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( rejectionLevel, argv[i+1] );
						    i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid rejection level: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;;
							}
							else if( rejectionLevel <= 0.0 || rejectionLevel >= 1.0 ) {
								errorCount++;
								cerr << "Error: The rejection level must be within the range (0,1).\n"
								     << "       You provided: " << rejectionLevel << "\n" << endl;
							}
						}
					}
					break;

				// Calculate T* for left tail tests. Usage: -L
				case 'L':
					// don't complain if the -L option appears more than once
					LOptionProvided = true;
					if( ROptionProvided ) {
						errorCount++;
						cerr << "Error: You cannot calculate T* for left and right tail tests at the same time.\n";
					}
					break;

				// Calculate T* for right tail tests. Usage: -R
				case 'R':
					// don't complain if the -R option appears more than once
					ROptionProvided = true;
					if( LOptionProvided ) {
						errorCount++;
						cerr << "Error: You cannot calculate T* for left and right tail tests at the same time.\n";
					}
					break;

				// Output T* to TSTAR_FILENAME. Usage: -T TSTAR_FILENAME
				case 'T':
					if( TOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one output file for T*.\n" << endl;
					}
					else {
						TOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a T* output filename with the -T option.\n" << endl;
						}
						else {
							TStarFilename = argv[i+1];
							i++;
						}
					}
					break;

				// Output sequences used to calculate T* to a file.
				// Cannot be used with the -i option. Usage: -o FILENAME
				case 'o':
					if( oOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one sequence output file.\n" << endl;
					}
					else {
						oOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a sequence output filename with the -o option.\n" << endl;
						}
						else {
							sequenceOutputFilename = argv[i+1];
							i++;
						}
					}
					break;

				// Output the T histogram to HISTOGRAM_FILENAME. Usage: -H HISTOGRAM_FILENAME
				case 'H':
					if( HOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one T histogram output file.\n" << endl;
					}
					else {
						HOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a T histogram output filename with the -H option.\n" << endl;
						}
						else {
							histogramFilename = argv[i+1];
							i++;
						}
					}
					break;

				// Use a total of BINS bins when generating the histogram. Usage -b BINS
				case 'b':
					if( bOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one total number of histogram bins.\n" << endl;
					}
					else {
						bOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a number of histogram bins with the -b option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( numberOfBins, argv[i+1] );
							i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid number of histogram bins: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;
							}
							else if( numberOfBins < 1 ) {
								errorCount++;
								cerr << "Error: The number of histogram bins must be at least 1.\n"
								     << "       You provided: " << numberOfBins << "\n" << endl;
							}
						}
					}
					break;

				// Use SMOOTHING_PARAMETER when detecting the peaks in G
				// (the default is 0.005). Usage: -s SMOOTHING_PARAMETER
				case 's':
					if( sOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one smoothing parameter.\n" << endl;
					}
					else {
						sOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a smoothing parameter with the -s option.\n" << endl;
						}
						else {
							conversionOutcome = stringToNumber( smoothingParameter, argv[i+1] );
						    i++;

							if( conversionOutcome != SUCCESS ) {
								errorCount++;
								cerr << "Error: Invalid smoothing parameter: ";
								printStringToNumberError( conversionOutcome );
								cerr << "       You provided: " << argv[i] << "\n" << endl;;
							}
							else if( smoothingParameter < 0 ) {
								errorCount++;
								cerr << "Error: The smoothing parameter must be positive semi-definite.\n"
								     << "       You provided: " << smoothingParameter << "\n" << endl;
							}
						}
					}
					break;

				// Use DELIMITER to separate fields in the input and output files
				// (the default is a comma). Usage: -d DELIMITER
				case 'd':
					if( dOptionProvided ) {
						errorCount++;
						cerr << "Error: You requested more than one delimiter.\n" << endl;
					}
					else {
						dOptionProvided = true;

						if( i+1 >= argc ) {
							errorCount++;
							cerr << "Error: You didn't provide a delimiter with the -d option.\n" << endl;
						}
						else {
							if( strlen(argv[i+1]) > 1 ) {
								errorCount++;
								cerr << "Error: The delimiter can only be a single character.\n"
								     << "       You provided: " << argv[i+1] << "\n" << endl;
							}
							else {
								delimiter = argv[i+1][0];
								i++;
							}
						}
					}
					break;

				// Print the version number to stderr
				case 'V':
					if( not VOptionProvided ) {
						printVersion();
						VOptionProvided = true;
					}
					break;

				default: // unknown option
					errorCount++;
					cerr << "Error: Unknown option: " << argv[i] << "\n" << endl;
			}
		}
		else {  // unknown option (the argument had more than 1 character)
			errorCount++;
			cerr << "Error: Unknown option: " << argv[i] << "\n" << endl;
		}
	}

	// The user only wanted the version number
	if( VOptionProvided && argc == 2 ) {
		return 0;
	}

	// Check if we received all the mandatory arguments

	if( iOptionProvided ) {
		if( nOptionProvided ) {
			errorCount++;
			cerr << "Error: You cannot specify the number of events per sequence when reading\n"
			     << "       sequences from a file.\n" << endl;
		}

		if( NOptionProvided ) {
			errorCount++;
			cerr << "Error: You cannot specify the sequence length when reading sequences from\n"
			     << "       a file.\n" << endl;
		}

		if( nOptionProvided ) {
			errorCount++;
			cerr << "Error: You cannot specify the total number of sequences when reading sequences\n"
			        "       from a file.\n" << endl;
		}

		if( oOptionProvided ) {
			errorCount++;
			cerr << "Error: You cannot write the sequences to a file when reading them from a file\n" << endl;
		}
	}
	else {
		if( not nOptionProvided ) {
			errorCount++;
			cerr << "Error: Either the number of events per sequence (-n option) or a sequence input file (-i option) is required.\n" << endl;
		}

		if( not NOptionProvided ) {
			errorCount++;
			cerr << "Error: Either the sequence length (-N option) or a sequence input file (-i option) is required.\n" << endl;
		}

		if( not rOptionProvided ) {
			errorCount++;
			cerr << "Error: Either the number of sequences (-r option) or a sequence input file (-i option) is required.\n" << endl;
		}
	}

	if( not cOptionProvided ) {
		errorCount++;
		cerr << "Error: The rejection level is required.\n" << endl;
	}

	// Bail if there were any errors processing the arguments
	if( errorCount > 0 ) {
		bail( argv[0], errorCount );
	}

	// Check if the arguments are valid

	if( nOptionProvided && NOptionProvided && numberOfEvents > sequenceLength ) {
		errorCount++;
		cerr << "Error: The number of events per sequence must be less than or equal to the sequence length.\n"
		     << "       You provided:\n"
		     << "           n = " << numberOfEvents << "\n"
             << "           N = " << sequenceLength << "\n" << endl;
	}

	// Bail if the arguments were not valid
	if( errorCount > 0 ) {
		bail( argv[0], errorCount );
	}

	if( iOptionProvided ) {
	// Import the sequences
		if( not importSequencesFromFile( sequenceInputFilename, sequences ) ) {
			bail( argv[0], ++errorCount );
		}
		sequenceInputFile.close();

		numberOfSequences = sequences.size();
		sequenceLength = sequences[0].N;
		numberOfEvents = sequences[0].events.size();

		// Check if our imported sequences are consistent
		// (all have the same length and number of events)
		for( unsigned int i = 0; i < numberOfSequences; i++ ) {
			if( sequences[i].N != sequenceLength ) {
				errorCount++;
				cerr << "Error: Sequence number " << i+1 << " has " << sequences[i].N << " comparison sites but was expecting " << sequenceLength << ".\n" << endl;
			}
			if( sequences[i].events.size() != numberOfEvents ) {
				errorCount++;
				cerr << "Error: Sequence number " << i+1 << " has " << sequences[i].events.size() << " events but was expecting " << numberOfEvents << ".\n" << endl;
			}
		}

		// bail if our data is inconsistent
		if( errorCount > 0 ) {
			bail( argv[0], errorCount );
		}
	}
	else {
	// Generate random sequences
		seed_drand48();

		for( unsigned int i = 0; i < numberOfSequences; i++ ) {
			eventSequence seq;
			seq.events = generateUniqueRandomUnsignedInts( 1, sequenceLength, numberOfEvents );
			seq.N = sequenceLength;
			sequences.push_back( seq );
		}
	}

	// Try to open the output files

	if( TOptionProvided ) {
		if( not openOutputFile( TStarFile, TStarFilename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	if( oOptionProvided ) {
		if( not openOutputFile( sequenceOutputFile, sequenceOutputFilename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	if( HOptionProvided ) {
		if( not openOutputFile( histogramFile, histogramFilename ) ) {
			bail( argv[0], ++errorCount );
		}
	}

	// calculate the T values

	for( unsigned int i = 0; i < numberOfSequences; i++ ) {
		vector<double> g = G( sequences[i] );
		Ts.push_back( T( g, smoothingParameter ) );
	}

	// Histogram the T values

	if( not bOptionProvided ) {
		numberOfBins = floor( sqrt( sequenceLength * numberOfEvents ) );
	}

	histogram = THistogram( Ts, numberOfBins );

	// Find T*

	if( LOptionProvided ) {
		tStar = oneTailTest( histogram, rejectionLevel, true );
	}
	else if( ROptionProvided ) {
		tStar = oneTailTest( histogram, rejectionLevel, false );
	}
	else {
		tStar = twoTailTest( histogram, rejectionLevel );
	}

	// Output the data

	if( TOptionProvided ) {
		TStarFile.precision(precision);
		TStarFile << tStar << endl;
		TStarFile.close();
	}
	else {
		cout << tStar << endl;
	}

	if( oOptionProvided ) {
		printVector( sequences, sequenceOutputFile, '\n' );
		sequenceOutputFile << "\n";
		sequenceOutputFile.close();
	}

	if( HOptionProvided ) {
		histogramFile.precision(precision);
		printVector( histogram, histogramFile, delimiter );
		histogramFile.close();
	}

	return 0;
}
